/*
 * Copyright (c) 2011 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: rht.c,v 1.8 2014/01/30 16:07:22 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/queue.h"
#include "sm/mta.h"
#include "sm/io.h"
#include "sm/actdb-int.h"
#include "adb.h"

/*	Note: locking must be provided by caller. */

/*
required functionality:

aq_rcpt_find_ss( sessta_id_T ss_ta_id, rcpt_idx_T rcpt_idx
aq_rcpt_find_one_ss( sessta_id_T ss_ta_id

*/

/*
**  AQ_RHT_INIT -- allocate and initialize hash
**
**	Parameters:
**		aq_ctx -- AQ context
**		htsize -- size of hash aq_ctx
**
**	Returns:
**		usual sm_error code; ENOMEM
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
aq_rht_init(aq_ctx_P aq_ctx, uint htsize)
{
	SM_IS_AQ(aq_ctx);
	aq_ctx->aq_rhtdata = (aq_rhd_T *) sm_zalloc(htsize * sizeof(aq_rhd_P));
	if (NULL == aq_ctx->aq_rhtdata)
		return sm_error_temp(SM_EM_HT, ENOMEM);
	aq_ctx->aq_rhtsize = htsize;
	aq_ctx->aq_rhtused = 0;
	return SM_SUCCESS;
}

/*
**  AQ_RHT_DESTROY -- destroy hash
**
**	Parameters:
**		aq_ctx -- AQ context
**
**	Returns:
**		Nothing.
**
**	Last code review:
**	Last code change:
*/

void
aq_rht_destroy(aq_ctx_P aq_ctx)
{
	SM_IS_AQ(aq_ctx);
	if (aq_ctx->aq_rhtdata != NULL) {
		sm_free((void *) aq_ctx->aq_rhtdata);
		aq_ctx->aq_rhtdata = NULL;
	}
}

/* Check whether two keys x and y are equal */
#define AQ_RHT_KEY_EQ(x, y, l) ((x)[0] == (y)[0] && sm_memeq((x), (y), l))

static uint
b16_uint(const char *key, uint b, uint e)
{
	ulong h;
	uint off;
	int c;

	for (off = b, h = 0; off <= e; off++) {
		c = (uchar)key[off];
		if (isascii(c) && isdigit(c))
			c -= '0';
		else if (isascii(c) && isalpha(c))
			c -= isupper(c) ? 'A' - 10 : 'a' - 10;
		h <<= 4;
		h += c;
	}
	return h;
}

/*
**  AQ_RHT_HASH_S -- hash a RCPT key
**
**	Parameters:
**		key -- key to hash
**
**	Returns:
**		hash value
**
**	Last code review:
**	Last code change:
*/

/* ss_ta_id */
static uint
aq_rht_hash_s(const char *key, rcpt_idx_T rcpt_idx)
{
	ulong h;

	h = b16_uint(key, 9, 16) + rcpt_idx;
	return h;
}

#define AQ_RHT_LINK(aq_ctx, aqrhd, aq_rcpt, hash, link) do {				\
		aqrhd = aq_ctx->aq_rhtdata + hash % aq_ctx->aq_rhtsize;	\
		RHTINCRACOL(aq_ctx, aqrhd); \
		LIST_INSERT_HEAD(aqrhd, aq_rcpt, link);	\
		RHTCHAINLEN(aq_ctx, aqrhd, aq_rcpt);	\
		aq_ctx->aq_rhtused++;				\
	} while (0)

/*
**  AQ_RHT_ADD -- add aq_rcpt
**	Note: this does not check whether key exists! It will simply add
**		the new pair.
**
**	Parameters:
**		aq_ctx -- AQ context
**		aq_rcpt -- aq_rcpt
**
**	Returns:
**		usual sm_error code; SM_E_FULL
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
aq_rht_add(aq_ctx_P aq_ctx, aq_rcpt_P aq_rcpt)
{
	uint hash;
	aq_rhd_P aqrhd;

	SM_IS_AQ(aq_ctx);
	SM_IS_AQ_RCPT(aq_rcpt);

	if (aq_ctx->aq_rhtused >= aq_ctx->aq_rhtsize)
		return sm_error_temp(SM_EM_HT, SM_E_FULL);
	hash = aq_rht_hash_s(aq_rcpt->aqr_ss_ta_id, 0);
	AQ_RHT_LINK(aq_ctx, aqrhd, aq_rcpt, hash, aqr_htl_s);
	return SM_SUCCESS;
}

/*
**  AQ_RHT_FIND_S -- lookup value
**
**	Parameters:
**		aq_ctx -- AQ context
**		ta_id -- key to hash
**		rcpt_idx -- recipient index
**
**	Returns:
**		pointer to value
**
**	Last code review:
**	Last code change:
*/

aq_rcpt_P
aq_rht_find_s(aq_ctx_P aq_ctx, sessta_id_T ta_id, rcpt_idx_T rcpt_idx)
{
	aq_rhd_P aqrhd;
	aq_rcpt_P aq_rcpt, aq_rcpt_ret;

	SM_IS_AQ(aq_ctx);
	aqrhd = aq_ctx->aq_rhtdata + aq_rht_hash_s(ta_id, 0) % aq_ctx->aq_rhtsize;
	aq_rhtincrlu(aq_ctx);
	aq_rhtinitchain(aq_ctx);
	aq_rcpt_ret = NULL;
	LIST_FOREACH(aq_rcpt, aqrhd, aqr_htl_s) {
		SM_IS_AQ_RCPT(aq_rcpt);
		if (AQ_RHT_KEY_EQ(ta_id, aq_rcpt->aqr_ss_ta_id, SMTP_STID_SIZE)) {
			if (aq_rcpt->aqr_idx != rcpt_idx)
				return aq_rcpt;
			aq_rcpt_ret = aq_rcpt;
		}
		aq_rhtincrlcol(aq_ctx);
		RHTINCRCHAIN(aq_ctx);
	}
	return aq_rcpt_ret;
}

/*
**  AQ_RHT_FIND_R -- lookup value
**
**	Parameters:
**		aq_ctx -- AQ context
**		ta_id -- key to hash
**		rcpt_idx -- rcpt_idx
**
**	Returns:
**		pointer to value
**
**	Last code review:
**	Last code change:
*/

aq_rcpt_P
aq_rht_find_r(aq_ctx_P aq_ctx, sessta_id_T ta_id, rcpt_idx_T rcpt_idx)
{
	aq_rcpt_P aq_rcpt;
	aq_rhd_P aqrhd;

	SM_IS_AQ(aq_ctx);
	aqrhd = aq_ctx->aq_rhtdata + aq_rht_hash_s(ta_id, 0) % aq_ctx->aq_rhtsize;
	aq_rhtincrlu(aq_ctx);
	aq_rhtinitchain(aq_ctx);
	LIST_FOREACH(aq_rcpt, aqrhd, aqr_htl_s) {
		SM_IS_AQ_RCPT(aq_rcpt);
		if (AQ_RHT_KEY_EQ(ta_id, aq_rcpt->aqr_ss_ta_id, SMTP_STID_SIZE)
		   && rcpt_idx == aq_rcpt->aqr_idx)
		{
			return aq_rcpt;
		}
		aq_rhtincrlcol(aq_ctx);
		RHTINCRCHAIN(aq_ctx);
	}
	return NULL;
}

/*
**  AQ_RHT_RM_R -- remove first entry that matches ta_id
**
**	Parameters:
**		aq_ctx -- AQ context
**		ta_id -- ta_id to look up
**		rcpt_idx -- rcpt_idx
**
**	Returns:
**		usual
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
aq_rht_rm_r(aq_ctx_P aq_ctx, sessta_id_T ta_id, rcpt_idx_T rcpt_idx)
{
	aq_rhd_P aqrhd;
	aq_rcpt_P aq_rcpt;

	SM_IS_AQ(aq_ctx);
	aqrhd = aq_ctx->aq_rhtdata + aq_rht_hash_s(ta_id, 0) % aq_ctx->aq_rhtsize;
	LIST_FOREACH(aq_rcpt, aqrhd, aqr_htl_s) {
		SM_IS_AQ_RCPT(aq_rcpt);
		if (AQ_RHT_KEY_EQ(ta_id, aq_rcpt->aqr_ss_ta_id, SMTP_STID_SIZE)
		   && rcpt_idx == aq_rcpt->aqr_idx)
		{
			LIST_REMOVE(aq_rcpt, aqr_htl_s);
			SM_ASSERT(aq_ctx->aq_rhtused > 0);
			aq_ctx->aq_rhtused--;
			return SM_SUCCESS;
		}
	}
	return SM_E_NOTFOUND;
}


void
aq_rht_stats(aq_ctx_P aq_ctx, char *out, uint len)
{
	SM_IS_AQ(aq_ctx);
	SM_REQUIRE(out != NULL);
	SM_REQUIRE(len > 0);
#if SM_HT_PERF
	sm_snprintf(out, len, "adds              %9lu\n"
		"lookups           %9lu\n"
		"deletes           %9lu\n"
		"add collisions    %9lu\n"
		"lookup collisions %9lu\n"
		"delete collisions %9lu\n"
		"max chain len     %9u\n"
		"grows             %9u\n"
		, aq_ctx->aq_rhtadds, aq_ctx->aq_rhtlookups, aq_ctx->aq_rhtdeletes
		, aq_ctx->aq_rhtacollisions
		, aq_ctx->aq_rhtlcollisions, aq_ctx->aq_rhtdcollisions
		, aq_ctx->aq_rhtchainmax, aq_ctx->aq_rhtgrows);
#else /* SM_HT_PERF */
	*out = '\0';
#endif /* SM_HT_PERF */
}
